//@ts-check
"use strict";

/**
 * @typedef NPCName
 * @type {keyof typeof npcDescs}
 */

/**
 * @typedef NPCDesc
 * @property {NPCName} id
 * @property {boolean} [uncounted]
 */

/**
@type {{
  grandma: NPCDesc,
  farmhand: NPCDesc,
  jack: NPCDesc,
  blacksmith: NPCDesc,
  knightR: NPCDesc,
  knightG: NPCDesc,
  knightY: NPCDesc,
  knightB: NPCDesc,
  mom: NPCDesc,
  bro: NPCDesc,
  hermit: NPCDesc,
  sheep: NPCDesc,
  wizard: NPCDesc,
  bard: NPCDesc,
  conan: NPCDesc,
  broody: NPCDesc,
  gudul: NPCDesc,
  pup: NPCDesc,
  samurai: NPCDesc,
  captain: NPCDesc,
  onion: NPCDesc,
  leek: NPCDesc,
  salad: NPCDesc,
}}
 */
const npcDescs = {
  grandma: {
    id: "grandma",
  },
  farmhand: {
    id: "farmhand",
  },
  jack: {
    id: "jack",
  },
  blacksmith: {
    id: "blacksmith",
  },
  knightR: {
    id: "knightR",
  },
  knightG: {
    id: "knightG",
  },
  knightY: {
    id: "knightY",
  },
  knightB: {
    id: "knightB",
  },
  mom: {
    id: "mom",
  },
  bro: {
    id: "bro",
  },
  hermit: {
    id: "hermit",
  },
  sheep: {
    id: "sheep",
  },
  wizard: {
    id: "wizard",
  },
  bard: {
    id: "bard",
  },
  conan: {
    id: "conan",
  },
  broody: {
    id: "broody",
  },
  gudul: {
    id: "gudul",
  },
  pup: {
    id: "pup",
  },
  samurai: {
    id: "samurai",
  },
  captain: {
    id: "captain",
  },
  onion: {
    id: "onion",
    uncounted: true,
  },
  salad: {
    id: "salad",
    uncounted: true,
  },
  leek: {
    id: "leek",
    uncounted: true,
  },
}

/**
 * @enum {number}
 */
const NpcStates = {
  Idle: 0,
  ChaseTarget: 1,
  Hit: 2
};

/**
 * @typedef NPC
 * @type {{
     id: string,
     desc: NPCDesc,
     pos: PIXI.Point,
     collideBox: { width: number; height: number };

     speed: number;
     health: number;
     state: NpcStates;
     stateTimer: number;

     knockbackDirection: PIXI.Point;

     pixiRoot: PIXI.Container,
     sprite: PIXI.AnimatedSprite,
     spritesheet: Spritesheet,
     shadowSprite: PIXI.Sprite,
     outlineFilter: PIXI.filters.OutlineFilter;

     following: boolean;
 * }}
 */

/**
 * @param {number} dt
 */
function updateNPCs(dt) {
  const { baxton } = inGame;
  const { hovered } = inGame.stats;

  for (let npc of inGame.activeWorld.npcs) {
    let isHovered = (hovered && hovered.kind == "npc" && hovered.id == npc.id);
    npc.outlineFilter.thickness = (isHovered ? 1 : 0) * app.renderer.height / gameHeight;

    if (!npc.following) continue;

    if (npc.state === NpcStates.Hit) {
      npc.stateTimer -= dt;
      if (npc.stateTimer <= 0) {
        npc.sprite.loop = true;
        setNpcIdle(npc);
      } else {
        moveWithCollision(npc, npc.knockbackDirection);
        npc.knockbackDirection.x *= 0.9;
        npc.knockbackDirection.y *= 0.9;
      }
    }

    if (npc.desc.id === "salad") {
      let foundBaddieNearby = false;

      for (let baddie of inGame.activeWorld.baddies) {
        const dist = distxy(baddie.pos, npc.pos);
        if (dist < 8) {
          foundBaddieNearby = true;
          break;
        }
      }

      if (foundBaddieNearby) {
        setNpcIdle(npc);
        continue;
      }
    }

    // Common follow behavior
    const diff = { x: baxton.pos.x - npc.pos.x, y: baxton.pos.y - npc.pos.y };
    const angle = Math.atan2(diff.y, diff.x);
    const dist = lengthxy(diff);

    const tooFarRange = 15;
    const sleepRange = 2;

    switch (npc.state) {
      case NpcStates.Idle: {
        npc.stateTimer -= dt;

        if (npc.stateTimer <= 0 && dist > sleepRange && dist < tooFarRange) {
          setNpcChaseTarget(npc, baxton);
        }
        break;
      }
      case NpcStates.ChaseTarget: {
        if (dist <= sleepRange) {
          setNpcIdle(npc);

          // TODO: Try to attack if chasing an enemy
          // setNpcAttack(npc);
        } else if (dist >= tooFarRange) {
          // TODO: Try to force a way towards the player
          setNpcIdle(npc);
        } else {
          const amount = Math.min(dist - npc.collideBox.width + baxton.collideBox.width, npc.speed * dt);
          moveWithCollision(npc, { x: Math.cos(angle) * amount, y: Math.sin(angle) * amount });
        }

        break;
      }
    }
  }

  for (const npc of inGame.activeWorld.npcs) {
    for (const otherNpc of inGame.activeWorld.npcs) {
      if (npc !== otherNpc && npc.following && otherNpc.following) {
        pushaway(npc, otherNpc, dt);
      }
    }

    if (npc.following) {
      // baxton pushes npcs away
      pushaway(npc, inGame.baxton, dt, false);
    } else {
      // non-follower npcs push baxton away
      pushaway(inGame.baxton, npc, dt, false);
    }
  }

  for (const npc of inGame.activeWorld.npcs) {
    npc.pixiRoot.position.copyFrom(npc.pos);
    npc.pixiRoot.zIndex = npc.pos.y;

    npc.shadowSprite.position.copyFrom(npc.pos);
    npc.shadowSprite.zIndex = npc.pos.y - 10;

    npc.sprite.update(app.ticker.deltaTime);
  }
}

/**
 * @param {World} world
 * @param {PointLike} pos
 * @param {NPCDesc} desc
 */
function createNPC(world, pos, desc, following = false) {
  const pixiRoot = new PIXI.Container();
  pixiRoot.scale.set(inverseTileSize);
  pixiRoot.position.set(pos.x, pos.y);
  world.pixiLayers["ground"].entitiesContainer.addChild(pixiRoot);

  const spritesheet = getSpritesheet(`inGame/npcs/${desc.id}/body`);

  const sprite = new PIXI.AnimatedSprite(spritesheet.animations["idle"], false);
  pixiRoot.addChild(sprite);
  sprite.anchor.copyFrom(spritesheet.anchor);
  sprite.play();
  const outlineFilter = new PIXI.filters.OutlineFilter(0, 0xffffff, 0.5);
  sprite.filters = [outlineFilter];

  const shadowSprite = new PIXI.Sprite(tex("inGame/shadow"));
  shadowSprite.scale.set(inverseTileSize * 1.2);
  shadowSprite.anchor.set(0.5, 0);
  shadowSprite.position.set(pos.x, pos.y);
  world.pixiShadowsLayer.addChild(shadowSprite);

  /** @type {NPC} */
  const npc = {
    id: generateID(),
    desc,
    pos: new PIXI.Point(pos.x, pos.y),
    collideBox: { width: 0.8, height: 0.45 },

    speed: 4,
    health: 2,
    state: NpcStates.Idle,
    stateTimer: 0,
    knockbackDirection: new PIXI.Point(),

    pixiRoot,
    sprite,
    spritesheet,
    shadowSprite,
    outlineFilter,

    following
  };

  world.npcs.push(npc);
}

/**
 * @param {NPC} npc
 */
function setNpcIdle(npc) {
  npc.state = NpcStates.Idle;
  if (npc.desc.uncounted) {
    // vegetable npcs get tired
    npc.stateTimer = 2;
  } else {
    npc.stateTimer = .7;
  }

  setAnimation(npc.sprite, npc.spritesheet.animations.idle);
}

/**
 * @param {NPC} npc
 * @param {EntityLike} target
 */
function setNpcChaseTarget(npc, target) {
  npc.state = NpcStates.ChaseTarget;

  setAnimation(npc.sprite, npc.spritesheet.animations.run);
}

/**
 * @param {NPC} npc
 * @param {number} angle
 */
function setNpcHit(npc, angle) {
  npc.state = NpcStates.Hit;
  npc.stateTimer = 0.3;

  npc.knockbackDirection.set(-Math.cos(angle) * 0.05, -Math.sin(angle) * 0.05);

  npc.sprite.loop = false;
  setAnimation(npc.sprite, npc.spritesheet.animations.hit);
}

/**
 * @param {NPC} npc
 * @param {number} damage
 * @param {number} angle
 */
function damageNpc(npc, damage, angle) {
  npc.health = Math.max(npc.health - damage, 0);

  play2Dsound(`${npc.health === 0 ? "deaths" : "hits"}.npcs.${npc.desc.id}`, npc.pos);

  if (npc.health === 0) {
    destroyNpc(npc);
    spawnAnimatedFx("inGame/fxs/vanish", npc.pos, 0);
  } else {
    setNpcHit(npc, angle);
  }
}

/**
 * @param {NPC} npc
 */
function destroyNpc(npc) {
  inGame.activeWorld.pixiLayers["ground"].entitiesContainer.removeChild(npc.pixiRoot);
  inGame.activeWorld.pixiShadowsLayer.removeChild(npc.shadowSprite);
  inGame.activeWorld.npcs.splice(inGame.activeWorld.npcs.indexOf(npc), 1);

  npc.sprite.destroy();
  npc.shadowSprite.destroy();
}
